/*=========================================================================*/
/**
	@file		wpputimg.c
	@brief		wizpy image put utility
	@version	0.02
	@Date
	@author
/*
/*=========================================================================*/

#include	<stdio.h>

#define	SCREEN_WIDTH	128				// wizpy screen width & height
#define	SCREEN_HEIGHT	160
#define	SCREEN_BPP		32

										// bit15:0 bit14-10:R bit9-5:G bit4-0:B
#define	WIZPY_MASK_R	0x7C00			// 0b0111_1100_0000_0000
#define	WIZPY_MASK_G	0x03E0			// 0b0000_0011_1110_0000
#define	WIZPY_MASK_B	0x001F			// 0b0000_0000_0001_1111

#define	ADDRESS_IMAGE_START		0x0054DF26
#define	ADDRESS_IMAGE_END		0x009E6F56


/* BitmapFileHeader */
typedef struct BITMAPFILEHEADER_tag {
	unsigned short bfType;
	unsigned int bfSize;
	unsigned short bfReserved1;
	unsigned short bfReserved2;
	unsigned int bfOffBits;
} BitmapFileHeader;

/* BitmapInfoHeader */
typedef struct BITMAPINFOHEADER_tag {
	unsigned int biSize;
	int biWidth;
	int biHeight;
	unsigned short biPlanes;
	unsigned short biBitCount;
	unsigned int biCompression;
	unsigned int biSizeImage;
	int biXPixPerMeter;
	int biYPixPerMeter;
	unsigned int biClrUsed;
	unsigned int biClrImportant;
} BitmapInfoHeader;

BitmapFileHeader BmpFileHeader;
BitmapInfoHeader BmpInfoHeader;

unsigned int address = 0x00;			// wizpy image data address
unsigned int width = 128;				// image width
unsigned int height = 160;				// image height

										// wizpy image data buffer
unsigned char wizbuf[ SCREEN_WIDTH * SCREEN_HEIGHT * 2 ];
										// BMP(32bpp) image data buffer
unsigned char bmpbuf[ SCREEN_WIDTH * SCREEN_HEIGHT * 4 ];



/**
	@fn			int isValidAddress( unsigned int adrs )
	@brief		check valid address or not
	@param		adrs : target address
	@return		TRUE : valid address
	@return		FALSE : invalid address ( out of image store area )
 */
int isValidAddress( unsigned int adrs )
{
	return ( ( adrs >= ADDRESS_IMAGE_START ) && ( adrs <= ADDRESS_IMAGE_END ) );
}


/**
	@fn			int readBMP( const char *filename )
	@brief		read BMP file
	@param		*filename : BMP filename
	@return		BMP filesize
 */
int readBMP( const char *filename )
{
	FILE *fp = NULL;

	/* open BMP file */
	fp = fopen( filename, "rb" );
	if( fp == NULL ) {
		printf( "file open failed : %s.\n", filename );
		return 0;
	}

	/* get BitmapFileHeader */
	fread( &BmpFileHeader.bfType, sizeof( BmpFileHeader.bfType ), 1, fp );
	fread( &BmpFileHeader.bfSize, sizeof( BmpFileHeader.bfSize ), 1, fp );
	fread( &BmpFileHeader.bfReserved1, sizeof( BmpFileHeader.bfReserved1 ), 1, fp );
	fread( &BmpFileHeader.bfReserved2, sizeof( BmpFileHeader.bfReserved2 ), 1, fp );
	fread( &BmpFileHeader.bfOffBits, sizeof( BmpFileHeader.bfOffBits ), 1, fp );

	/* get BitmapInfoHeader */
	fread( &BmpInfoHeader.biSize, sizeof( BmpInfoHeader.biSize ), 1, fp );
	fread( &BmpInfoHeader.biWidth, sizeof( BmpInfoHeader.biWidth ), 1, fp );
	fread( &BmpInfoHeader.biHeight, sizeof( BmpInfoHeader.biHeight ), 1, fp );
	fread( &BmpInfoHeader.biPlanes, sizeof( BmpInfoHeader.biPlanes ), 1, fp );
	fread( &BmpInfoHeader.biBitCount, sizeof( BmpInfoHeader.biBitCount ), 1, fp );
	fread( &BmpInfoHeader.biCompression, sizeof( BmpInfoHeader.biCompression ), 1, fp );
	fread( &BmpInfoHeader.biSizeImage, sizeof( BmpInfoHeader.biSizeImage ), 1, fp );
	fread( &BmpInfoHeader.biXPixPerMeter, sizeof( BmpInfoHeader.biXPixPerMeter ), 1, fp );
	fread( &BmpInfoHeader.biYPixPerMeter, sizeof( BmpInfoHeader.biYPixPerMeter ), 1, fp );
	fread( &BmpInfoHeader.biClrUsed, sizeof( BmpInfoHeader.biClrUsed ), 1, fp );
	fread( &BmpInfoHeader.biClrImportant, sizeof( BmpInfoHeader.biClrImportant ), 1, fp );

	/* read Bitmap data */
	fread( bmpbuf, BmpInfoHeader.biSizeImage, 1, fp );

	/* close file */
	fclose( fp );

	return BmpInfoHeader.biSizeImage;
}


/**
	@fn			void writeWizData( FILE *fwp )
	@brief		write wizpy 16bpp image to firmware
	@param		*fwp : firmware file pointer
	@return		null
 */
void writeWizData( FILE *fwp )
{
	/* seek to image address */
	if( fseek( fwp, address, SEEK_SET ) != 0 ) {
		printf( "firmware file seek failed.\n" );
		return;
	}

	/* write bitmap data */
	fwrite( wizbuf, ( width * height * 2 ), 1, fwp );

	return;
}


/**
	@fn			void conv24to16( void )
	@brief		convert 24bpp image(bmpbuf) to wizpy 16bpp image(wizbuf)
	@param		null
	@return		null
 */
void conv24to16( void )
{
	int x = 0;
	int y = 0;
	int rp = 0;
	int wp = 0;
	unsigned short color16 = 0;
	int adjust = ( width % 4 );			// adjust bytes (4-byte boundary for each scan line)

	width = BmpInfoHeader.biWidth;
	height = BmpInfoHeader.biHeight;

	for( y = 0; y < height; y++ ) {
		for( x = 0; x < width; x++ ) {
			/* read and write pointer */
			wp = ( x + ( y * width ) ) * 2;
			rp = ( x * 3 ) + ( ( height - y - 1 ) * ( width * 3 + adjust ) );

			/* convert 32bpp to 16bpp */
			color16 = 0;
			color16 |= (unsigned short)( ( bmpbuf[ rp + 2 ] >> 3 ) << 10 );
			color16 |= (unsigned short)( ( bmpbuf[ rp + 1 ] >> 3 ) <<  5 );
			color16 |= (unsigned short)( bmpbuf[ rp ] >> 3 );
			wizbuf[ wp + 1 ] = (unsigned char)( color16 >> 8 );
			wizbuf[ wp ] = (unsigned char)( color16 );
		}
	}

	return;
}


/**
	@fn			void conv32to16( void )
	@brief		convert 32bpp image(bmpbuf) to wizpy 16bpp image(wizbuf)
	@param		null
	@return		null
 */
void conv32to16( void )
{
	int x = 0;
	int y = 0;
	int rp = 0;
	int wp = 0;
	unsigned short color16 = 0;

	width = BmpInfoHeader.biWidth;
	height = BmpInfoHeader.biHeight;

	for( y = 0; y < height; y++ ) {
		for( x = 0; x < width; x++ ) {
			/* read and write pointer */
			wp = ( x + ( y * width ) ) * 2;
			rp = ( x + ( ( height - y - 1 ) * width ) ) * 4;

			/* convert 32bpp to 16bpp */
			color16 = 0;
			color16 |= (unsigned short)( ( bmpbuf[ rp + 2 ] >> 3 ) << 10 );
			color16 |= (unsigned short)( ( bmpbuf[ rp + 1 ] >> 3 ) <<  5 );
			color16 |= (unsigned short)( bmpbuf[ rp ] >> 3 );
			wizbuf[ wp + 1 ] = (unsigned char)( color16 >> 8 );
			wizbuf[ wp ] = (unsigned char)( color16 );
		}
	}

	return;
}


/**
	@fn			void printUsage( void )
	@brief		print usage
	@param		null
	@return		null
 */
void printUsage( void )
{
	printf( "wizpy firmware image put utility  ver 0.2\n" );
	printf( "Syntax: wpputimg <firmware-file> <address(hex)> <width> <height> <input-file(.bmp)>\n" );
	printf( "        or\n" );
	printf( "        wpputimg <firmware-file> -l <listfile>\n" );
	printf( "\n" );

	return;
}


/**
	@fn			int main( int argc, char *argv[] )
	@brief		main routine
	@param		argc
	@param		*argv
	@return
 */
int main( int argc, char *argv[] )
{
	FILE *fpfirm;						// firmware file
	FILE *fplist;						// list file
	int mode = 0;
	int i = 0;
	char listarg[ 256 ];
	char namearg[ 128 ];
	char filename[ 128 + 4 ];
	unsigned int size;


	/* check options */
	if( argc < 2 ) {
		printUsage();
		return 1;
	}
	/* check batch mode */
	else if( argv[2][0] == '-' ) {
		if( argv[2][1] == 'l' ) {
			mode = 1;
		}
	}
	else {
		if( argc < 6 ) {
			printUsage();
			return 1;
		}
	}

	/* open firmware file */
	fpfirm = fopen( argv[1], "rb+" );
	if( fpfirm == NULL ) {
		printf( "firmware file open failed.\n" );
		return 1;
	}


	/* batch mode */
	if( mode == 1 ) {
		/* open list file */
		fplist = fopen( argv[3], "rt" );
		if( fplist == NULL ) {
			printf( "list file open failed.\n" );
			return 1;
		}

		/* read list file */
		while( fgets( listarg, sizeof( listarg ), fplist ) != NULL ) {
			if( listarg[ 0 ] != '#' ) {
				/* get parameters */
				sscanf( listarg, "0x%08X 0x%04X %3d %3d %s", &address, &size, &width, &height, &namearg );

				/* read BMP file */
				sprintf( filename, "%s.bmp", namearg );
				printf("address:0x%08X, width:%d, height:%d, output:%s\n", address, width, height, filename );
				if( isValidAddress( address ) ) {
					if( readBMP( filename ) > 0 ) {
						/* check bitcount */
						if( BmpInfoHeader.biBitCount == 32 ) {
							/* convert BMP(32bpp) to wizpy image */
							conv32to16();
							/* write wizpy image data */
							writeWizData( fpfirm );
						}
						else if( BmpInfoHeader.biBitCount == 24 ) {
							/* convert BMP(24bpp) to wizpy image */
							conv24to16();
							/* write wizpy image data */
							writeWizData( fpfirm );
						}
						else {
							printf( "unsupported bitcount.\n" );
						}
					}
				}
				else {
					printf( "invalid address.\n" );
				}
			}
		}

		/* close list file */
		if( fclose( fplist ) != 0 ) {
			printf( "list file close failed.\n" );
		}
	}
	/* single mode */
	else {
		/* get parameters */
		sscanf( argv[2], "%x", &address );
		sscanf( argv[3], "%d", &width );
		if( width > SCREEN_WIDTH ) {
			printf( "parameter error : width\n" );
		}
		sscanf( argv[4], "%d", &height );
		if( height > SCREEN_HEIGHT ) {
			printf( "parameter error : height\n" );
		}
		printf("address:0x%08X, width:%d, height:%d, output:%s\n", address, width, height, argv[5] );

		/* check address */
		if( isValidAddress( address ) ) {
			/* read BMP file */
			if( readBMP( argv[5] ) > 0 ) {
				/* check bitcount */
				if( BmpInfoHeader.biBitCount == 32 ) {
					/* convert BMP(32bpp) to wizpy image */
					conv32to16();
					/* write wizpy image data */
					writeWizData( fpfirm );
				}
				else if( BmpInfoHeader.biBitCount == 24 ) {
					/* convert BMP(24bpp) to wizpy image */
					conv24to16();
					/* write wizpy image data */
					writeWizData( fpfirm );
				}
				else {
					printf( "unsupported bitcount.\n" );
				}
			}
		}
		else {
			printf( "invalid address.\n" );
		}
	}

	/* close firmware file */
	if( fclose( fpfirm ) != 0 ) {
		printf( "input file close failed.\n" );
	}

	return 0;
}
